/**
  *  \file client/dialogs/consoledlg.cc
  */

#include "client/dialogs/consoledlg.h"
#include "client/widgets/consoleview.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets.h"
#include "ui/window.h"
#include "util/console.h"
#include "util/nls.h"
#include "int/bytecode.h"
#include "int/exec.h"
#include "int/lex.h"
#include "int/expr.h"
#include "int/parseexpr.h"
#include "int/interror.h"
#include "int/intglobal.h"

namespace {
    class WConsoleDialog : public UIWindow {
        ObjectHolder h;
        UIInputLine* input;
        WConsoleView* cv;
        int recall_index;
        string_t recall_last_line;

        void init();
     public:
        WConsoleDialog();
        void onEnter();
        void doRecall(int direction);
        void doInsert();
        bool handleEvent(const UIEvent& e, bool second_pass);
    };
}

WConsoleDialog::WConsoleDialog()
    : UIWindow(_("Console"), UIWindow::BLUE_WINDOW, COLOR_GRAY), recall_index(-1)
{
    init();
}

void
WConsoleDialog::init()
{
    // VBox
    //   WConsoleViewWithHandler
    //   UIInputLine
    //   HBox
    //     UIButton "Evaluate"
    //     UIButton "Close"
    //     UISpacer
    //    [UIButton "Help"]
    setLayout(UIVBoxLayout::instance5);

    UIGroup& buttons = h.add(new UIGroup(UIHBoxLayout::instance5));

    cv = new WConsoleView(console, 10);
    add(h.add(cv));
    input = new UIInputLine(INT_MAX, 30, 0, UIInputLine::ilf_Framed, 0);
    add(h.add(input));
    add(buttons);

    UIButton& button_eval  = h.add(new UIButton(0, _("Evaluate"), SDLK_RETURN));
    UIButton& button_close = h.add(new UIButton(0, _("Close"), SDLK_ESCAPE));
    buttons.add(button_eval);
    buttons.add(button_close);
    buttons.add(h.add(new UISpacer()));

    button_close.addStop();
    button_eval.addCall(this, &WConsoleDialog::onEnter);

    input->setFocusedWidget();

    pack();
}

/** Evaluate entered expression. */
void
WConsoleDialog::onEnter()
{
    IntBCORef bco = new IntBytecodeObject();

    // Reset recall
    recall_index = -1;

    // Compile
    try {
        IntTokenizer tok(input->getText());
        if (tok.getCurrentToken() == tok.tEnd) {
            // blank input; ignore silently
            return;
        }

        // Parse it
        std::auto_ptr<IntExprNode> expr(IntExpressionParser(tok).parse());
        if (tok.getCurrentToken() != tok.tEnd)
            throw IntError("Expression incorrectly terminated (missing operator?)");

        // Compile it
        expr->compileValue(*bco, IntCompilationContext());

        // Run it
        IntExecutionContext exec;
        exec.pushContextsFrom(global_contexts);
        exec.pushNewFrame(new IntExecutionFrame(bco));
        exec.run();

        // Check result
        IntValue* result = exec.getResult();
        console.write(LOG_INFO, input->getText(), ConsoleRecallable);
        if (result != 0)
            console.write(LOG_INFO, result->toString(true), ConsoleRightJustified | ConsoleRecallable);
        else
            console.write(LOG_INFO, "Empty", UIColor::tc_Faded | ConsoleRightJustified);

        // Clear input line for next expression
        input->setText("");
    }
    catch (IntError& e) {
        console.write(LOG_ERROR, e.what());
    }
    catch (std::exception& e) {
        console.write(LOG_ERROR, e.what());
    }
}

/** "Recall operation": browse through history of previous recallable input and output. */
void
WConsoleDialog::doRecall(int direction)
{
    int index = recall_index;
    if (index < 0)
        /* Remember final input */
        recall_last_line = input->getText();
    while (1) {
        index += direction;
        if (index >= console.getScrollbackFill()) {
            /* We cannot scroll so far back into the past */
            break;
        }
        if (index < 0) {
            /* We have reached the present */
            if (recall_index >= 0) {
                /* We were in the past */
                input->setText(recall_last_line);
            }
            recall_index = -1;
            break;
        }
        if (console.getLine(index).format & ConsoleRecallable) {
            /* We have found a message that can be recalled */
            recall_index = index;
            input->setText(console.getLine(index).text);
            break;
        }
    }
}

/** "Insert" operation: insert last result into current input. */
void
WConsoleDialog::doInsert()
{
    for (int index = 0; index < console.getScrollbackFill(); ++index) {
        if (console.getLine(index).format == (ConsoleRightJustified | ConsoleRecallable)) {
            /* This is our last result */
            input->setSelectedWidget();
            input->insert(console.getLine(index).text);
            break;
        }
    }
}

bool
WConsoleDialog::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case ss_Shift + SDLK_PAGEUP:
            cv->setSelectedWidget();
            cv->scrollBack(5);
            return true;
         case ss_Shift + SDLK_UP:
            cv->setSelectedWidget();
            cv->scrollBack(1);
            return true;
         case ss_Shift + SDLK_PAGEDOWN:
            cv->setSelectedWidget();
            cv->scrollBack(-5);
            return true;
         case ss_Shift + SDLK_DOWN:
            cv->setSelectedWidget();
            cv->scrollBack(-1);
            return true;
         case SDLK_UP:
            doRecall(+1);
            return true;
         case SDLK_DOWN:
            doRecall(-1);
            return true;
         case SDLK_INSERT:
            doInsert();
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

void
doConsole()
{
    WConsoleDialog win;
    {
        const char* message = _("Enter command or expression:");
        if (console.getScrollbackFill() == 0 || console.getLine(0).text != message)
            console.write(LOG_INFO, message);
    }
    ui_root->centerWidget(win);
    ui_root->runModal(win);
}
