/**
  *  \file client/scr-planet.cc
  *  \brief Planet Screen
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/actions.h"
#include "client/cscreen.h"
#include "client/dlg-list.h"
#include "client/marks.h"
#include "client/screens.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/objlist.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/font.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "util/nls.h"
#include "client/dlg-tech.h"
#include "ui/layout.h"
#include "client/dlg-structure.h"
#include "client/dlg-tax.h"
#include "ui/table.h"
#include "client/dlg-supplies.h"
#include "resmgr/resid.h"
#include "ui/util.h"
#include "game/tables.h"
#include "client/wscanchart.h"
#include "int/intglobal.h"
#include "int/data.h"
#include "client/dlg-comment.h"
#include "game/config.h"
#include "client/tiles/planetecon.h"
#include "client/tiles/planetnatives.h"
#include "client/tiles/planetcolonist.h"
#include "client/tiles/planetfcode.h"
#include "client/tiles/planetbase.h"
#include "client/dlg-objsel.h"
#include "client/tiles/planetoverview.h"
#include "game/host.h"
#include "int/if/planetif.h"
#include "client/dialogs/consoledlg.h"

namespace {
    class WPlanetScreen : public WControlScreen {
     public:
        WPlanetScreen();
        ~WPlanetScreen();
        void update();
        bool handleEvent(const UIEvent& e, bool second_pass);
    };

    class WPlanetScreenHeaderTile : public WControlScreenHeaderTile {
     public:
        WPlanetScreenHeaderTile(GObjectType& type);
        string_t getTitle();
        string_t getSubtitle();
        ResId getPictureId();
    };
}

/** Choose a planet. Displays the dialog if necessary and lets the
    user choose a planet.
    \return planet Id, zero if no planet or cancelled */
int
choosePlanet()
{
    int nplanets = global_planet_type.getCount();
    if (nplanets == 0) {
        messageBox(_("You do not have any planets. "
                     "Unload colonists from a starship to a planet to colonize it."),
                   _("Planets"));
        return 0;
    } else if (nplanets == 1) {
        return global_planet_type.getCurrent();
    } else {
        GAliasObjectType aot(global_planet_type);
        WPlanetOverviewTile bot(aot);
        WObjectSelectionDialog dlg(_("Select Planet"), aot);
        dlg.addWidget(bot);
        dlg.pack();
        ui_root->centerWidget(dlg);
        if (ui_root->runModal(dlg) == cm_Enter)
            return aot.getCurrent();
        else
            return 0;
    }
}

/***************************** WPlanetScreen *****************************/

/** \class WPlanetScreen
    \brief Planet Control Screen

    Standard control screen for a player's planet. */

/** Constructor. */
WPlanetScreen::WPlanetScreen()
    : WControlScreen(global_planet_type, WScreenId::PlanetScreen)
{
    addTile(holder.add(new WPlanetScreenHeaderTile(global_planet_type)));
    addTile(holder.add(new WPlanetEconomyTile(global_planet_type)),
            _("Economy:"));
    addTile(holder.add(new WPlanetNativeTile(global_planet_type)),
            _("Natives:"));
    addTile(holder.add(new WPlanetColonistTile(global_planet_type)),
            _("Colonists:"));
    addTile(holder.add(new WPlanetFCodeTile(global_planet_type)));
    addTile(holder.add(new WPlanetBaseTile(global_planet_type, false)));
    addChart(holder.add(new WScannerChartWidget()));
}

WPlanetScreen::~WPlanetScreen()
{ }

void
WPlanetScreen::update()
{
    if (GMapObject* mo = dynamic_cast<GMapObject*>(&type.getObject(type.getCurrent())))
        getScanner().getChart()->setCenterPosition(mo->getPos());
}

static void
doPlanetXchg(WPlanetScreen& scr)
{
    WScanner& scanner = scr.getScanner();
    int pid = scanner.getScannedPlanetId(scanner.getScannedShipId() != 0);
    if (pid && pid != scr.getType().getCurrent() && scr.getType().isExistant(pid))
        scr.stopWithScreenId(WScreenId(WScreenId::PlanetScreen, pid));
    else
        putKey(ss_Ctrl + SDLK_F1);
}

bool
WPlanetScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        try {
            const int pid = getType().getCurrent();
            GUniverse& univ = *getType().getUniverse();
            GPlanet& planet = univ.getPlanet(pid);
            switch(e.key.code) {
             case 'x':
                doPlanetXchg(*this);
                return true;
             case 'c':
             case 'u':
                doPlanetTransfer(univ, planet, e.key.code == 'u');
                return true;
             case 'f':
                doPlanetFCodeChange(univ, planet);
                return true;
             case 'b':
             case SDLK_TAB:
                doPlanetBuildScreen(univ, planet, 0);
                return true;
             case 'm':
                doPlanetBuildScreen(univ, planet, 1);
                return true;
             case 'd':
                doPlanetBuildScreen(univ, planet, 2);
                return true;
             case 't':
                doTaxation(univ, planet);
                return true;
             case 's':
                WSellSuppliesDialog::doDialog(univ, planet);
                return true;
             case 'g':
                if (host.isPHost() && config.CPEnableGive())
                    doGivePlanet(planet, getDisplayedTurn().getCommands(planet.getOwner()));
                univ.doScreenUpdates();
                return true;
             case ss_Ctrl + 's':
                WUndoSellSuppliesDialog::doDialog(getDisplayedTurn() /* FIXME */, pid);
                return true;
             case SDLK_F8:
                if (planet.hasBase()) {
                    stopWithScreenId(WScreenId(WScreenId::BaseScreen, pid));
                } else {
                    doPlanetBuildBase(univ, planet);
                }
                return true;
             case SDLK_F9:
                doEditComment(planet, getPlanetProperties(planet.getId()), _("Enter new comment for this planet:"), _("Edit Comment"));
                return true;
             case SDLK_INSERT:
                doPlanetAddToSim(planet);
                return true;
             case ss_Alt + 'c':
                {
                    IntTemporaryContext con(new IntPlanetContext(planet.getId()));
                    doConsole();
                }
                return true;
                /* planet screen keys:
                   i        mail
                   alt-r    reset */
            }
        }
        catch (const GError& e) {
            messageBox(e.getUserError(), _("Error")); // FIXME
            return true;
        }
    }
    return WControlScreen::handleEvent(e, second_pass);
}

WPlanetScreenHeaderTile::WPlanetScreenHeaderTile(GObjectType& type)
    : WControlScreenHeaderTile(type, bAuto + bAdd + bSend)
{ }

string_t
WPlanetScreenHeaderTile::getTitle()
{
    return getObject()->getName();
}

string_t
WPlanetScreenHeaderTile::getSubtitle()
{
    string_t result;
    if (GPlanet* p = dynamic_cast<GPlanet*>(getObject())) {
        int16 level = p->unit_scores.getScore(planet_score_definitions.lookupScore(ScoreId_ExpLevel));
        const char* fmt = (level >= 0
                           ? _("(Id #%d, %s - %dF, %s)")
                           : _("(Id #%d, %s - %dF)"));
        result = format(fmt,
                        p->getId(),
                        getTemperatureName(p->getTemperature()),
                        p->getTemperature(),
                        getExperienceLevelName(level));
    }
    return result;
}

ResId
WPlanetScreenHeaderTile::getPictureId()
{
    ResId result;
    if (GPlanet* p = dynamic_cast<GPlanet*>(getObject()))
        result = ResId(res::planet, p->getTemperature(), p->getId());
    return result;
}

WScreenId
doPlanetScreen()
{
    return WPlanetScreen().runScreen();
}
