/**
  *  \file client/scr-ship.cc
  *  \brief Starship Control Screen
  *
  *  This contains the ship screen tile widgets, and the main driver
  *  code.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/actions.h"
#include "client/cscreen.h"
#include "client/dlg-cargo.h"
#include "client/dlg-comment.h"
#include "client/dlg-fcode.h"
#include "client/dlg-list.h"
#include "client/dlg-mission.h"
#include "client/dlg-warp.h"
#include "client/marks.h"
#include "client/screens.h"
#include "client/wcolorframe.h"
#include "client/wscanchart.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/fcode.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/host.h"
#include "game/mission.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/spec.h"
#include "game/univ.h"
#include "gfx/clip.h"
#include "gfx/palette.h"
#include "int/data.h"
#include "int/intglobal.h"
#include "resmgr/resid.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "util/misc.h"
#include "util/nls.h"
#include "client/tiles/shipmission.h"
#include "client/tiles/shipequipment.h"
#include "client/tiles/shipmovement.h"
#include "client/tiles/shipcargo.h"
#include "client/dlg-objsel.h"
#include "client/tiles/shipoverview.h"
#include "game/team.h"
#include "game/racename.h"
#include "client/remote.h"
#include "game/shipquery.h"
#include "client/widgets/specview.h"
#include "int/if/shipif.h"
#include "client/dialogs/consoledlg.h"

namespace {
    /// Ship scanner (with waypoint etc).
    class WShipScannerChartWidget : public WScannerChartWidget {
     protected:
        void  lockQueryLocation(bool& hyperjumping, int& current_gravity_pid);
        int32 lockQueryDistance(GPoint pt);
        
     public:
        WShipScannerChartWidget();
        ~WShipScannerChartWidget();
        void drawScanner(GfxCanvas& can);
        void drawPost(GfxCanvas& can);
        // void update(int);
        void userMoveScannerTo(GPoint pt);
    };

    /// Starship screen.
    class WShipScreen : public WControlScreen {
        WShipScannerChartWidget scanner_chart;

        void doShipBaseMenu(GShip& ship, UIBaseWidget& button);
     public:
        WShipScreen();
        ~WShipScreen();
        bool handleEvent(const UIEvent& e, bool second_pass);
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        void update();
        void setWaypointToScanner();
    };

    class WShipScreenHeaderTile : public WControlScreenHeaderTile {
     public:
        WShipScreenHeaderTile(GObjectType& type);
        string_t getTitle();
        string_t getSubtitle();
        ResId getPictureId();
    };
}

static int acquireWaypointPerm(const string_t& reason, GShip& ship);

WTile*
createShipScreenHeaderTile(GObjectType& type)
{
    return new WShipScreenHeaderTile(type);
}

int
chooseShip()
{
    int nships = global_ship_type.getCount();
    if (nships == 0) {
        messageBox(_("You do not have any ships. "
                     "To build ships, use [B] on a starbase."),
                   _("Starships"));
        return 0;
    } else if (nships == 1) {
        return global_ship_type.getCurrent();
    } else {
        GAliasObjectType aot(global_ship_type);
        WShipOverviewTile bot(aot);
        WObjectSelectionDialog dlg(_("Select Ship"), aot);
        dlg.addWidget(bot);
        dlg.pack();
        ui_root->centerWidget(dlg);
        if (ui_root->runModal(dlg) == cm_Enter)
            return aot.getCurrent();
        else
            return 0;
    }
}

/************************ WShipScannerChartWidget ************************/

WShipScannerChartWidget::WShipScannerChartWidget()
{ }

WShipScannerChartWidget::~WShipScannerChartWidget()
{ }

void
WShipScannerChartWidget::lockQueryLocation(bool& hyperjumping, int& current_gravity_pid)
{
    const GShip& sh = global_ship_type.getShip();
    hyperjumping        = sh.isHyperdriving();
    current_gravity_pid = getDisplayedTurn().getCurrentUniverse().getGravityPlanetAt(sh.getPos());
}

int32
WShipScannerChartWidget::lockQueryDistance(GPoint pt)
{
    const GShip& sh = global_ship_type.getShip();
    const int32 dist2 = distanceSquared(sh.getPos(), pt);
    if (sh.isHyperdriving() && !host.isExactHyperjumpDistance2(dist2)) {
        return -1;
    } else {
        return dist2;
    }
}

void
WShipScannerChartWidget::drawScanner(GfxCanvas& can)
{
    const GChartViewport& vp = getViewport();

    GPoint                wp = global_ship_type.getShip().getWaypoint();
    GfxPoint       wp_screen = vp.scaleXY(wp);

    if (wp != getScannerPosition() && wp_screen != vp.getCenter()) {
        GfxParameters parm;
        parm.color = standard_colors[COLOR_RED];

        drawLine(can, parm, vp.getCenter(), wp_screen);
        drawEndBig(can, parm, wp_screen);
    }

    WScannerChartWidget::drawScanner(can);
}

void
WShipScannerChartWidget::drawPost(GfxCanvas& can)
{
    // Draw mine laying info
    GMineMissionData md;
    if (parseMineLayingMission(global_ship_type.getShip(), getDisplayedTurn().getCurrentUniverse(), md)) {
        // We're laying a minefield. What kind?
        GMinefield* mf = 0;
        if (md.mine_id > 0)
            mf = getDisplayedTurn().getCurrentUniverse().getMinefields().get(md.mine_id);

        // Figure out where it goes
        GPoint center;
        int radius;
        string_t notice;
        if (mf == 0) {
            // New minefield
            center = global_ship_type.getShip().getPos();
            radius = GMinefield::getRadiusFromUnits(md.num_units);
            notice = format(_("%d ly"), radius);
        } else {
            // Enlarge existing minefield
            center = mf->getPos();
            radius = GMinefield::getRadiusFromUnits(md.num_units + mf->getUnitsForLaying());
            notice = format(_("%d ly (+%d)"), radius, radius - mf->getRadius());
        }

        // Drawing parameters
        GfxParameters parm;
        const GChartViewport& vp = getViewport();
        GfxClipRectFilter can1(getExtent(), can);
        GfxPoint p = vp.scaleXY(center.getNearestTo(vp.center));

        parm.color  = standard_colors[COLOR_YELLOW];
        parm.just_x = 1;
        parm.just_y = 0;
        parm.font   = fonts[FONT_SMALL];

        // Limit check
        int limit = (md.type
                     ? config.MaximumWebMinefieldRadius(md.owner)
                     : config.MaximumMinefieldRadius(md.owner));
        if (radius > limit) {
            drawCircle(can1, parm, p, vp.scale(limit));
            parm.color = standard_colors[COLOR_DARK];
            outText(can1, parm, p.x, p.y + 10 + font_heights[FONT_SMALL], _("<over limit>"));
        }

        // Draw circle
        drawCircle(can1, parm, p, vp.scale(radius));
        outText(can1, parm, p.x, p.y + 10, notice);
    } else if (parseMineScoopMission(global_ship_type.getShip(), getDisplayedTurn().getCurrentUniverse(), md)) {
        // We're scooping
        const GMinefieldContainer& mfc = getDisplayedTurn().getCurrentUniverse().getMinefields();
        const GShip& sh = global_ship_type.getShip();
        int room = sh.getFreeCargo();
        for (GMinefieldContainer::const_iterator mf = mfc.begin(); mf != mfc.end(); ++mf) {
            // Check whether we can scoop this field
            if (mf->getOwner() == md.owner
                && distanceSquared(mf->getPos(), sh.getPos()) <= mf->getUnitsForLaying()
                && (md.mine_id == 0 || md.mine_id == mf->getId()))
            {
                // Okay, scoop it. First, figure out conversion rate.
                int rate = ((mf->isWeb() ? config.UnitsPerWebRate(md.owner) : config.UnitsPerTorpRate(md.owner))
                            * square(sh.getTorpType()))
                    / 100;
                if (rate == 0)
                    rate = 1;

                // How much do we clear?
                int32 units = mf->getUnitsForLaying();
                int32 scoopedUnits;
                int scoopedTorps = units / rate;
                if (scoopedTorps > room) {
                    // We cannot clear this field
                    scoopedTorps = room;
                    scoopedUnits = room * rate;
                } else {
                    // We can clear this field
                    scoopedUnits = units;
                }

                // Check limit
                if (md.num_torps > 0 && md.num_torps < scoopedTorps) {
                    scoopedTorps = md.num_torps;
                    scoopedUnits = md.num_torps * rate;
                }

                // Draw it
                if (scoopedTorps > 0) {
                    GfxParameters parm;
                    const GChartViewport& vp = getViewport();
                    GfxClipRectFilter can1(getExtent(), can);
                    GfxPoint p = vp.scaleXY(mf->getPos().getNearestTo(vp.center));

                    parm.color  = standard_colors[COLOR_YELLOW];
                    parm.just_x = 1;
                    parm.just_y = 0;
                    parm.font   = fonts[FONT_SMALL];
                
                    room -= scoopedTorps;
                    if (units > scoopedUnits) {
                        // Field stays
                        int radius = mf->getRadiusFromUnits(units - scoopedUnits);
                        drawCircle(can1, parm, p, radius);
                        outText(can1, parm, p.x, p.y + 10, format(_("%d ly (%d torp%!1{s%})"), radius, scoopedTorps));
                    } else {
                        // Field is gone
                        outText(can1, parm, p.x, p.y + 10, format(_("gone (%d torp%!1{s%})"), scoopedTorps));
                    }
                }
            }
        }
    }

    // Base class
    WScannerChartWidget::drawPost(can);
}

// void
// WShipScannerChartWidget::update(int)
// {
//     drawWidget(false);
// }

// void
// WShipScannerChartWidget::handleScannerChange(bool dblclick)
// {
//     if (dblclick) {
//         GGameTurn& trn = getDisplayedTurn();
//         if (int id = acquireWaypointPerm(_("waypoint"), trn.getShipType().getShip())) {
//             trn.getShip(id).setWaypoint(getScannerPosition());
//             trn.getCurrentUniverse().doUpdates();
//         }
//     }
// }

void
WShipScannerChartWidget::userMoveScannerTo(GPoint pt)
{
    bool scanner_tied_to_waypoint = false;
    if (scanner_tied_to_waypoint) {
        // FIXME
        setScannerPosition(pt);
    } else {
        setScannerPosition(pt);
    }
}

/****************************** WShipScreen ******************************/

WShipScreen::WShipScreen()
    : WControlScreen(global_ship_type, WScreenId::ShipScreen),
      scanner_chart()
{
    addTile(holder.add(new WShipScreenHeaderTile(global_ship_type)));
    addTile(holder.add(new WShipEquipmentTile(global_ship_type)),    _("Equipment & Crew:"));
    addTile(holder.add(new WShipCargoTile(global_ship_type)),        _("Aboard:"));
    addTile(holder.add(new WShipMissionTile(global_ship_type)),      _("Mission:"));
    addTile(holder.add(new WShipMovementTile(global_ship_type)),     _("Movement:"));
    addChart(scanner_chart);

    scanner_chart.sig_dblclick.add(this, &WShipScreen::setWaypointToScanner);
}

WShipScreen::~WShipScreen()
{ }

void
WShipScreen::update()
{
    int sid = getType().getCurrent();
    const GShip& ship = getType().getUniverse()->getShip(sid);
    scanner_chart.setCenterPosition(ship.getPos());
    scanner_chart.setScannerPosition(scanner_chart.clipScannerPosition(ship.getWaypoint()));
    scanner_chart.drawWidget();
}

static void
doShipXchg(WControlScreen& scr)
{
    WScanner& scan             = scr.getScanner();
    WScannerChartWidget* chart = scan.getChart();
    int n;
    if (scan.getScannedShipId()) {
        n = scr.getType().getUniverse()->getNumberOfPlayedShipsAt(scr.getScannerPosition());
        if (chart->getCenterPosition() == chart->getScannerPosition())
            --n;                // discount current ship
    } else {
        n = 0;
    }

    if (n)
        putKey(ss_Ctrl + SDLK_F1);
    else
        putKey(ss_Ctrl + SDLK_F2);
}

/** Attempt to acquire permission to change waypoint. (this is used for fleets)
    \param reason reason, for user
    \param ship   ship we want to change
    \returns 0 if permission not granted, ship id to change waypoint on
    specified ship */
static int
acquireWaypointPerm(const string_t& reason, GShip& ship)
{
    if (!ship.getWaypointPermission()) {
        // FIXME: display dialog, leave fleet, ...
        (void) reason;
        return 0;
    } else {
        return ship.getId();
    }
}

void
doShipTransferFor(GShip& ship, GObjectId chosen)
{
    GMultiTransaction mt;
    GShipTransfer&   left = mt.addNewSubtransaction(new GShipTransfer(ship));
    GCargoContainer* right;

    if (!chosen.isGameObject()) {
        /* jettison */
        right = &mt.addNewSubtransaction(new GShipTransporter(ship));
    } else if (GShip* s = dynamic_cast<GShip*>(&chosen.getObject())) {
        if (s->getOwner() == ship.getOwner()) {
            /* ship <-> ship */
            ASSERT(s->hasShipData());
            right = &mt.addNewSubtransaction(new GShipTransfer(*s));
        } else {
            /* ship <-> foreign ship */
            if (ship.isTransporterInUse(GShip::t_Transfer)) {
                messageBox(_("The cargo transporter is already in use."), _("Cargo Transfer"));
                return;
            }
            if (ship.getTransporter(GShip::t_Transfer).id != 0
                && ship.getTransporter(GShip::t_Transfer).id != s->getId()
                && ship.isTransporterActive(GShip::t_Transfer))
            {
                messageBox(format(_("The cargo transporter is already transferring cargo "
                                    "to ship #%d."),
                                  ship.getTransporter(GShip::t_Transfer).id), _("Cargo Transfer"));
                // FIXME: state name of current target ship here
                // FIXME: we could cancel the transfer here
                return;
            }
            // ship.cancelTranporter(true);
            right = &mt.addNewSubtransaction(new GShipTransporter(ship, GShip::t_Transfer, *s));
        }
    } else if (GPlanet* p = dynamic_cast<GPlanet*>(&chosen.getObject())) {
        if (p->getOwner() == ship.getOwner() && p->hasPlanetData()) {
            right = &mt.addNewSubtransaction(new GPlanetTransfer(*p));
        } else {
            if (ship.isTransporterInUse(GShip::t_Unload)) {
                messageBox(_("The cargo transporter is already in use."), _("Cargo Transfer"));
                return;
            }
            right = &mt.addNewSubtransaction(new GShipTransporter(ship, GShip::t_Unload, *p));
        }
    } else {
        FAIL();
        right = 0;
    }

    doCargoTransfer(mt, mt.addNewSubtransaction(new GCargoTransfer(left, *right)));
}

static void
doShipCargoTransfer(GUniverse& univ, GShip& ship)
{
    GObjectList list;
    int pid = univ.getPlanetAt(ship.getPos());
    if (pid)
        list.addObject(global_anyplanet_type, pid);
    else
        list.addNonObject(-1, _("Jettison into space"));

    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (i != ship.getId() && univ.getShip(i).isExistant() && univ.getShip(i).getPos() == ship.getPos())
            list.addObject(global_anyship_type, i);

    GObjectId chosen = WObjectSelectionList(0, 15, 320, list).doStandardDialog(_("Transfer cargo to..."), 0 /* FIXME */);
    if (chosen.isValid())
        doShipTransferFor(ship, chosen);
}

static void
doShipNavigationChart(GUniverse& univ, GShip& ship)
{
    int size = ui_root->getExtent().h;

    UIWindow the_window(_("Navigation Chart"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    the_window.setLayout(UIVBoxLayout::instance5);

    ObjectHolder h;

    WScannerChartWidget chart;
    chart.setCenterPosition(ship.getPos());
    chart.setScannerPosition(ship.getWaypoint());
    chart.setState(chart.sc_StyleChange | chart.sc_Zoomable | chart.sc_Keyboard, true);
    UIFrameGroup& fg = h.add(new UIFrameGroup(UIVBoxLayout::instance0));
    fg.add(chart);
    the_window.add(fg);

    UIGroup buttons(UIHBoxLayout::instance5);
    WScanner scanner;
    scanner.setChart(chart);
    buttons.add(scanner);
    buttons.add(h.add(new UISpacer()));
    buttons.add(h.add(new UIButton(cm_Enter, _("F10 - OK"), SDLK_F10)).addStop());
    buttons.add(h.add(new UIButton(cm_Escape, "ESC",        SDLK_ESCAPE)).addStop());
    the_window.add(buttons);
    the_window.setExtent(GfxRect(0, 0, size - 50, size - 10));

    ui_root->centerWidget(the_window);
    ui_root->add(the_window);
    int rv = ui_root->run();
    ui_root->close(the_window);

    if (rv == cm_Enter) {
        ship.setWaypoint(chart.getScannerPosition());
        setOptimumWarp(univ, ship.getId());
    }
}

void
WShipScreen::setWaypointToScanner()
{
    GGameTurn& trn = getDisplayedTurn();
    if (int id = acquireWaypointPerm(_("waypoint"), global_ship_type.getShip())) {
        trn.getShip(id).setWaypoint(getScannerPosition());
        setOptimumWarp(trn.getCurrentUniverse(), id);
        trn.getCurrentUniverse().doScreenUpdates();
    }
}

static void
doChangePE(GUniverse& univ, GShip& ship)
{
    UIStringList d;
    d.add(0, _("0 - none"));
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        d.add(i, format("%X - %s", i, player_racenames.getShortName(i)));

    UIStandardListbox lb(0, -1, 240, d);
    lb.scrollToKey(ship.getPE());
    if (lb.doStandardListbox(_("Change Primary Enemy"), 0))
        ship.setPE(lb.getKey());
    univ.doScreenUpdates();
}

void
WShipScreen::doShipBaseMenu(GShip& ship, UIBaseWidget& button)
{
    enum {
        CancelShipyard,
        FixThisShip,
        RecycleThisShip,
        GoToBase
    };

    // Are we orbiting a base?
    if (!global_base_type.isExistant(ship.getOrbit())) {
        messageBox(_("We are not at one of our starbases."), _("Starbase Commands"));
        return;
    }
    GPlanet& base = global_base_type.getObject(ship.getOrbit());
    GUniverse& univ = *global_base_type.getUniverse();

    // Build command menu
    UIStringList cmdlist;

    // - Fix/Recycle
    // FIXME: the "can we fix" logic is also in act-fix.cc; can we share it?
    int currentAction = base.getBaseShipyardId() == ship.getId() ? base.getBaseShipyardOrder() : 0;
    if (currentAction == sya_Fix) {
        cmdlist.add(CancelShipyard, _("Cancel \"repair\" order"));
    } else {
        cmdlist.add(FixThisShip, _("Fix (repair) this ship"));
    }
    if (currentAction == sya_Recycle) {
        cmdlist.add(CancelShipyard, _("Cancel \"recycle\" order"));
    } else {
        if (base.getOwner() == ship.getOwner())
            cmdlist.add(RecycleThisShip, _("Recycle this ship"));
    }

    // - Build torps/fighters FIXME

    // - Starbase screen
    cmdlist.add(GoToBase, _("Starbase screen"));

    // - Clone FIXME

    // Show menu
    UIStandardListbox lb(0, 1, getTextWidth(*fonts[FONT_NORMAL], cmdlist) + 20, cmdlist);
    if (!lb.doStandardMenu(button))
        return;

    switch (int cmd = lb.getKey()) {
     case CancelShipyard:
        base.setBaseShipyardOrder(0, 0);
        univ.doScreenUpdates();
        break;

     case FixThisShip:
     case RecycleThisShip:
        {
            bool accept = true;
            if (base.getBaseShipyardOrder() != sya_None && base.getBaseShipyardId() != ship.getId()) {
                // Find name of ship being repaired/recycled
                string_t shipName;
                if (univ.isValidShipId(base.getBaseShipyardId()) && univ.getShip(base.getBaseShipyardId()).isExistant())
                    shipName = univ.getShip(base.getBaseShipyardId()).getName();
                else
                    shipName = format(_("Ship %d"), base.getBaseShipyardId());
                // Ask user
                if (base.getBaseShipyardOrder() == sya_Fix) {
                    accept = UIMessageBox(format(_("This starbase is already repairing %s. "
                                                   "If you continue, this order will be replaced. Continue?"),
                                                 shipName),
                                          _("Starbase Commands")).doYesNoDialog();
                } else if (base.getBaseShipyardOrder() == sya_Recycle) {
                    accept = UIMessageBox(format(_("This starbase is already recycling %s. "
                                                   "If you continue, this order will be replaced. Continue?"),
                                                 shipName),
                                          _("Starbase Commands")).doYesNoDialog();
                }
            }
            if (accept) {
                base.setBaseShipyardOrder(cmd == FixThisShip ? sya_Fix : sya_Recycle,
                                          ship.getId());
                univ.doScreenUpdates();
            }
        }
        break;

     case GoToBase:
        stopWithScreenId(WScreenId(WScreenId::BaseScreen, ship.getOrbit()));
        break;
    }
}

bool
WShipScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        GUniverse& univ = *getType().getUniverse();
        GShip&     ship = global_ship_type.getShip();
        switch (e.key.code) {
         case 'n': {
             string_t s = ship.getName();
             if (UIInputLine::run(s, 20, format(_("Rename Starship #%d"),
                                                ship.getId()),
                             _("Enter new name:"), 20, 0))
                 ship.setName(s);
             univ.doScreenUpdates();
             return true;
         }
         case 'm':
            doShipMissionChange(univ, ship);
            return true;
         case 'e':
            doChangePE(univ, ship);
            return true;
         case 'a':
            if (int id = acquireWaypointPerm(_("waypoint"), ship)) {
                doShipNavigationChart(univ, univ.getShip(id));
                univ.doScreenUpdates();
            }
            return true;
         case 'c':
            doShipCargoTransfer(univ, ship);
            univ.doScreenUpdates();
            return true;
         case ss_Ctrl + 'j':
            if (univ.getPlanetAt(ship.getPos()) == 0) {
                doShipTransferFor(ship, GObjectId());
                univ.doScreenUpdates();
            }
            return true;
         case ss_Ctrl + 's':
            if (ship.getTransporter(GShip::t_Transfer).id) {
                doShipTransferFor(ship, GObjectId(global_anyship_type,
                                                  ship.getTransporter(GShip::t_Transfer).id));
                univ.doScreenUpdates();
            }
            return true;
         case ss_Ctrl + 'p':
            // FIXME: should possibly also handle beam-up orders
            if (ship.getTransporter(GShip::t_Unload).id) {
                doShipTransferFor(ship, GObjectId(global_anyplanet_type,
                                                  ship.getOrbit()));
                univ.doScreenUpdates();
            }
            return true;
         case 'u':
            if (ship.getOrbit()) {
                doShipTransferFor(ship, GObjectId(global_anyplanet_type,
                                                  ship.getOrbit()));
                univ.doScreenUpdates();
            } else {
                messageBox(_("We are not orbiting a planet, sir. "
                             "If you want to jettison cargo, use the "
                             "'c' command instead."),
                           _("Cargo Transfer"));
            }
            return true;
         case 'x':
            doShipXchg(*this);
            return true;
         case 'f': {
             GFCodeList list(GFCodeList::getInstance(), ship);
             WFCodeWindow dialog(_("Change Friendly Code"), list);
             dialog.setFCode(ship.getFCode());
             ui_root->centerWidget(dialog);
             ui_root->add(dialog);
             int rv = ui_root->run();
             ui_root->close(dialog);
             if (rv == cm_Enter) {
                 ship.setFCode(dialog.getFCode());
                 univ.doScreenUpdates();
             }
             return true;
         }
         case ss_Ctrl + 'f':
            if (ship.getOrbit()) {
                GPlanet& p = univ.getPlanet(ship.getOrbit());
                if (p.isFCodeKnown()) {
                    ship.setFCode(p.getFCode());
                    univ.doScreenUpdates();
                }
            }
            return true;
         case 'w':
            if (int id = acquireWaypointPerm(_("speed"), ship)) {
                (WShipSpeedSelector(univ.getShip(id)).doStandardDialog());
            }
            return true;
         case 'r':
            if (host.isPHost() && config.CPEnableRemote())
                doRemoteControl(ship, getDisplayedTurn().getCommands(ship.getOwner()));
            univ.doScreenUpdates();
            return true;
         case 'g':
            if (host.isPHost() && config.CPEnableGive())
                doGiveShip(ship, getDisplayedTurn().getCommands(ship.getOwner()));
            univ.doScreenUpdates();
            return true;
         case 's':
            {
                GShipQuery q;
                q.initForExistingShip(univ, ship.getId());
                WSpecView sv;
                sv.setQuery(q);
                sv.doStandardDialog();
            }
            return true;
         case ss_Ctrl + 'z':
            if (int id = acquireWaypointPerm(_("waypoint"), ship)) {
                ship.setWaypoint(ship.getPos());
                univ.doScreenUpdates();
            }
            return true;
         case ss_Ctrl + 'w':
            setWaypointToScanner();
            return true;
         case ss_Ctrl + 'i':
            // FIXME: if mission says arg is a planet, go to planet instead
            if (ship.getIntId())
                stopWithScreenId(WScreenId(WScreenId::ShipScreen, ship.getIntId()));
            return true;
         case ss_Ctrl + 't':
            // FIXME: if mission says arg is a planet, go to planet instead
            if (ship.getTowId())
                stopWithScreenId(WScreenId(WScreenId::ShipScreen, ship.getTowId()));
            return true;
         case SDLK_F8:
         case ss_Shift + SDLK_F8:
            stopWithScreenId(WScreenId(WScreenId::BaseScreen, ship.getOrbit()));
            return true;
         case SDLK_TAB:
         case ss_Shift + SDLK_TAB:
         case ss_Ctrl + SDLK_TAB:
         case ss_Shift + ss_Ctrl + SDLK_TAB: {
            GObjectList l;
            l.addObjectsAt(univ, ship.getPos(), 0, 0);
            l.postprocessUser(false);
            int cur = l.getIndexFor(ship);
            int n = ((e.key.code & ss_Shift)
                     ? l.getPreviousWrap(cur, (e.key.code & ss_Ctrl) != 0)
                     : l.getNextWrap    (cur, (e.key.code & ss_Ctrl) != 0));
            if (n > 0)
                stopWithScreenId(WScreenId(getScreenId(), l.getObject(n).getId()));
            return true;
         }
         case SDLK_F9:
            doEditComment(ship, getShipProperties(ship.getId()), _("Enter new comment for this ship:"), _("Edit Comment"));
            return true;
         case SDLK_INSERT:
            doShipAddToSim(ship, true);
            return true;
         case ss_Alt + 'c':
            {
                IntTemporaryContext con(new IntShipContext(ship.getId()));
                doConsole();
            }
            return true;

            /* ship screen keys:

               m-i, m-t         // search xceptor/tower
               c-n              // list next turn
               q                // toggle independantmove
               d                // distribute cargo
               f10              // fleet
               s-f10            // join fleet/goto fleet
               m-r              // reset x/y
            */
        }
    }
    return WControlScreen::handleEvent(e, second_pass);
}

bool
WShipScreen::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_Ship_BaseMenu) {
        doShipBaseMenu(global_ship_type.getShip(), sender);
        return true;
    } else {
        return WControlScreen::handleCommand(cmd, sender);
    }
}

WShipScreenHeaderTile::WShipScreenHeaderTile(GObjectType& type)
    : WControlScreenHeaderTile(type, bName + bAuto + bCycle + bAdd)
{ }

string_t
WShipScreenHeaderTile::getTitle()
{
    return getObject()->getName();
}

string_t
WShipScreenHeaderTile::getSubtitle()
{
    if (GShip* sh = dynamic_cast<GShip*>(getObject())) {
        /* Format is "(Id #%d, [race] [level] hull)".
           Since format() cannot easily suppress strings conditionally, we
           implement the suppression here. The suppressed parameters still
           appear in the format string, but the %! makes them not show up.
           This has the added benefit of making the format strings for the
           "(ID race hull)" and (ID level hull)" cases different which some
           languages may need. An alternative would have been to use the
           %nn$ syntax to move the argument pointer. */
        int16 level = sh->unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel));
        const char* fmt = (sh->getRealOwner() != getPlayerId()
                           ? (level >= 0
                              ? _("(Id #%d, %s %s %s)")
                              : _("(Id #%d, %s %!s%s)"))
                           : (level >= 0
                              ? _("(Id #%d, %!s%s %s)")
                              : _("(Id #%d, %!s%!s%s)")));
        return format(fmt)
            << sh->getId()
            << player_racenames.getAdjName(sh->getRealOwner())
            << getExperienceLevelName(level)
            << sh->getHull().getName();
    } else {
        return string_t();
    }
}

ResId
WShipScreenHeaderTile::getPictureId()
{
    if (GShip* sh = dynamic_cast<GShip*>(getObject())) {
        return ResId(res::ship, sh->getHull().getPicture(), sh->getHullId());
    } else {
        return ResId();
    }
}

WScreenId
doShipScreen()
{
    return WShipScreen().runScreen();
}
