/**
  *  \file int/bytecode.cc
  *  \brief Interpreter: Bytecode Object
  */

#include "int/bytecode.h"
#include "cpluslib/assert.h"
#include "int/intglobal.h"
#include "int/expr.h"           // IntCompilationContext

/** Construct blank BCO. */
IntBytecodeObject::IntBytecodeObject()
    : num_labels(0)
{ }

/** Destructor. */
IntBytecodeObject::~IntBytecodeObject()
{ }

/** Make a new label for future reference. This label can be used in as many jumps
    as needed (addJump), and must be placed exactly once using addLabel. */
IntBytecodeObject::label_t
IntBytecodeObject::makeLabel()
{
    return num_labels++;
}

/** Add an instruction. */
void
IntBytecodeObject::addInstruction(IntOpcode::Major major, uint8_t minor, uint16_t arg)
{
    IntOpcode o;
    o.major = major;
    o.minor = minor;
    o.arg   = arg;
    code.push_back(o);
}

/** Add variable-referencing instruction.
    \param major Major opcode
    \param name  Name of referenced variable
    \param cc    Compilation context */
void
IntBytecodeObject::addVariableReferenceInstruction(IntOpcode::Major major, const string_t& name, const IntCompilationContext& cc)
{
    if (cc.compilation_flags & IntCompilationContext::LocalContext) {
        /* Is it a local variable? */
        IntVariableNames::index_t ix = local_names.lookup(name);
        if (ix != local_names.nil) {
            addInstruction(major, IntOpcode::sLocal, ix);
            return;
        }

        /* Is it a global variable? */
        if (cc.compilation_flags & IntCompilationContext::AlsoGlobalContext) {
            ix = global_property_names.lookup(name);
            if (ix != local_names.nil) {
                addInstruction(major, IntOpcode::sShared, ix);
                return;
            }
        }
    }

    /* Shortcut not possible */
    addInstruction(major, IntOpcode::sNamedVariable, addName(name));
}

/** Place a label. */
void
IntBytecodeObject::addLabel(label_t label)
{
    addInstruction(IntOpcode::maJump, IntOpcode::jSymbolic, label);
}

/** Add a jump. */
void
IntBytecodeObject::addJump(uint8_t flags, label_t label)
{
    addInstruction(IntOpcode::maJump, flags | IntOpcode::jSymbolic, label);
}

/** Add a "push literal" instruction. Checks whether the literal can be
    encoded as a single instruction; otherwise generates a new literal table slot. */
void
IntBytecodeObject::addPushLiteral(IntValue* literal)
{
    /* Is it empty? */
    if (literal == 0) {
        addInstruction(IntOpcode::maPush, IntOpcode::sEmpty, 0);
        return;
    }

    /* Is it a small integer literal? */
    if (IntIntValue* iv = dynamic_cast<IntIntValue*>(literal)) {
        if (iv->getValue() >= -int32_t(0x7FFF) && iv->getValue() <= int32_t(0x7FFF)) {
            if (iv->isBool()) {
                addInstruction(IntOpcode::maPush, IntOpcode::sBoolean, iv->getValue());
            } else {
                addInstruction(IntOpcode::maPush, IntOpcode::sInteger, iv->getValue());
            }
            return;
        }
    }

    /* None of the above, so use general way */
    // FIXME: check 16-bit range
    // FIXME: recycle literals if possible
    data.push(literal);
    addInstruction(IntOpcode::maPush, IntOpcode::sLiteral, data.size()-1);
}

/** Add a name for reference by later instructions. Existing names are recycled if
    possible. */
uint16_t
IntBytecodeObject::addName(string_t name)
{
    return names.addMaybe(name);
}

/** Turn symbolic references into absolute references. */
void
IntBytecodeObject::relocate()
{
    // FIXME: can we degrade gracefully when there's more than 64k instructions?
    // FIXME: this leaves the LABEL null-operations in. We should compact the code
    // to remove them.

    std::vector<pc_t> addresses(num_labels, pc_t(0));

    // Find existing labels
    for (pc_t i = 0; i != code.size(); ++i) {
        if (code[i].major == IntOpcode::maJump && code[i].minor == (IntOpcode::jSymbolic + IntOpcode::jLabel)) {
            ASSERT(code[i].arg < addresses.size());
            addresses[code[i].arg] = i;
        }
    }

    // Turn symbolic jumps into absolute.
    for (pc_t i = 0; i != code.size(); ++i) {
        if (code[i].major == IntOpcode::maJump && (code[i].minor & IntOpcode::jSymbolic) != 0) {
            ASSERT(code[i].arg < addresses.size());
            code[i].arg = addresses[code[i].arg];
            code[i].minor &= ~IntOpcode::jSymbolic;
        }
    }
}

/** Find jump target. If the jump is symbolic, looks up the target label.
    \param minor Minor opcode from the maJump instruction
    \param arg Parameter */
IntBytecodeObject::pc_t
IntBytecodeObject::getJumpTarget(uint8_t minor, uint16_t arg)
{
    if (minor & IntOpcode::jSymbolic) {
        // Symbolic jump: find label
        for (pc_t i = 0; i != code.size(); ++i)
            if (code[i].major == IntOpcode::maJump && code[i].minor == (IntOpcode::jSymbolic + IntOpcode::jLabel))
                if (code[i].arg == arg)
                    return i;
        return code.size();
    } else {
        return arg;
    }
}

/** Get literal from literal table. */
IntValue*
IntBytecodeObject::getLiteral(uint16_t index)
{
    /* No range check required; it is done by IntDataSegment. */
    return data[index];
}

/** Get name from name table. */
string_t
IntBytecodeObject::getName(uint16_t index)
{
    return names.lookupAddress(index);
}

/** Get subroutine reference from subroutine table. */
IntBCORef
IntBytecodeObject::getSubroutine(uint16_t index)
{
    // FIXME: range check
    return subrs[index];
}
