/**
  *  \file int/compound.cc
  *  \brief Interpreter: Compound Values
  */

#include "int/compound.h"
#include "int/interror.h"
#include "int/context.h"

/************************ IntSimpleIndexableValue ************************/

/** Constructor.
    \param get  Function to implement get() (=indexed read access, function call).
                Can be null, in this case read access ('pushind' opcode) will not
                be permitted.
    \param make Function to implement makeFirstContext() (=iteration access).
                Can be null, in this case iteration ('firstindex' opcode) will not
                be permitted. */
IntSimpleIndexableValue::IntSimpleIndexableValue(get_t get, make_t make)
    : get_function(get), make_function(make)
{ }

IntSimpleIndexableValue::~IntSimpleIndexableValue()
{ }

// IntIndexableValue:
IntValue*
IntSimpleIndexableValue::get(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    if (get_function)
        return get_function(dseg, index, nargs);
    else
        throw IntError::typeError(IntError::tyIndexable);
}

void
IntSimpleIndexableValue::set(IntDataSegment& /*dseg*/, uint32_t /*index*/, uint32_t /*nargs*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntContext*
IntSimpleIndexableValue::makeFirstContext()
{
    if (make_function)
        return make_function();
    else
        throw IntError::typeError(IntError::tyIterable);
}

// IntValue:
string_t
IntSimpleIndexableValue::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntSimpleIndexableValue::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

IntValue*
IntSimpleIndexableValue::clone()
{
    return new IntSimpleIndexableValue(*this);
}

/************************* IntSimpleContextValue *************************/

/** Constructor.
    \param context Contained context. This object will hand out clones of that context.
                   This object will own the context. */
IntSimpleContextValue::IntSimpleContextValue(IntContext* context)
    : context(context)
{ }

IntSimpleContextValue::~IntSimpleContextValue()
{
    delete context;
}

// IntContextValue:
IntContext*
IntSimpleContextValue::makeContext()
{
    return context->clone();
}

// IntValue:
string_t
IntSimpleContextValue::toString(bool /*readable*/)
{
    return "#<context>";
}

void
IntSimpleContextValue::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

IntValue*
IntSimpleContextValue::clone()
{
    return createFromNewContext(context->clone());
}

/** Constructor. Implements exception-safe construction of
    IntSimpleContextValue. If memory doesn't suffice to create the
    IntSimpleContextValue, the IntContext is deleted as well. */
IntSimpleContextValue*
IntSimpleContextValue::createFromNewContext(IntContext* context)
{
    try {
        return new IntSimpleContextValue(context);
    }
    catch (...) {
        delete context;
        throw;
    }
}

/*************************** IntSubroutineValue **************************/

IntSubroutineValue::IntSubroutineValue(IntBCORef bco)
    : bco(bco)
{ }

IntSubroutineValue::~IntSubroutineValue()
{ }

IntBCORef
IntSubroutineValue::getBytecodeObject()
{
    return bco;
}

string_t
IntSubroutineValue::toString(bool /*readable*/)
{
    return "#<subr>";
}

void
IntSubroutineValue::store(IntTagNode& /*tag*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

IntSubroutineValue*
IntSubroutineValue::clone()
{
    return new IntSubroutineValue(bco);
}
