/**
  *  \file int/compound.h
  *  \brief Interpreter: Compound Values
  *
  *  This defines compound values, i.e. values which are not basic
  *  types of our language.
  *
  *  IntIndexableValue is the value which appears to the left of a
  *  function call "value(args)" or in iteration "ForEach value". 
  *  IntSimpleIndexableValue implements the common case of such a
  *  value, which is a global function with no associated data.
  *
  *  IntContextValue is a the value which appears to the left of a
  *  member reference expression "a .mem" (usually, "array(arg).mem"),
  *  or in "With a". It provides a way to obtain a context object from
  *  the value. IntSimpleContextValue implements the common case of
  *  such a value, which is an easily-clonable, lightweight context.
  *
  *  IntSubroutineValue is used to implement the 'pushsub' operation.
  *
  *  \todo Can we implement IntContext as child of IntValue, to avoid
  *  IntContextValue?
  *
  *  \todo Can we use IntSubroutineValue to actually implement user
  *  subroutines? How about global builtin commands? We probably need
  *  a middle-man, e.g. IntCommandValue, of which IntSubroutineValue
  *  and IntBuiltinCommandValue are derived.
  */
#ifndef PCC_V2_INT_COMPOUND_H
#define PCC_V2_INT_COMPOUND_H

#include "int/value.h"
#include "int/data.h"
#include "int/bytecode.h"

class IntContext;

/** Indexable value. This value can be used in two forms, "value(args)" and
    "ForEach value Do ...". */
class IntIndexableValue : public IntValue {
 public:
    /** Get indexed value. This implements "a := value(args)".
        \param dseg  Data segment containing parameters
        \param index Index of first parameter
        \param nargs Number of parameters
        \return obtained value. Throws IntError if request is invalid */
    virtual IntValue* get(IntDataSegment& dseg, uint32_t index, uint32_t nargs) = 0;

    /** Set indexed value. This implements "value(args) := a".
        \param dseg  Data segment containing parameters
        \param index Index of first parameter
        \param nargs Number of parameters
        \param value Value to assign.
        \return obtained value. Throws IntError if request is invalid */
    virtual void set(IntDataSegment& dseg, uint32_t index, uint32_t nargs, IntValue* value) = 0;

    /** Get context for first item in iteration. This should be equivalent to
        <tt>dynamic_cast<IntContextValue*>(get(args))->makeContext()</tt>, where
        args are the parameters needed to address the first object.
        Caller assumes lifetime management for the context.
        \return context. Return 0 if set is empty. Throws IntError if request is invalid. */
    virtual IntContext* makeFirstContext() = 0;
};

/** Simple indexable value. Implements an IntIndexableValue supporting read access
    and iteration, based upon two function pointers. In other words, this implements
    a global function, with possible iteration access. */
class IntSimpleIndexableValue : public IntIndexableValue {
 public:
    typedef IntValue*   (*get_t)(IntDataSegment&, uint32_t, uint32_t);
    typedef IntContext* (*make_t)();

    IntSimpleIndexableValue(get_t get, make_t make);
    ~IntSimpleIndexableValue();

    // IntIndexableValue:
    IntValue*   get(IntDataSegment& dseg, uint32_t index, uint32_t nargs);
    void        set(IntDataSegment& dseg, uint32_t index, uint32_t nargs, IntValue* value);
    IntContext* makeFirstContext();

    // IntValue:
    string_t    toString(bool readable);
    void        store(IntTagNode&, Stream&);
    IntValue*   clone();

 private:
    get_t  get_function;
    make_t make_function;
};

/** Context value. This value can be used as "value.member". Usually, this
    is the result of a function call, as in "func(x).member". */
class IntContextValue : public IntValue {
 public:
    /** Create the described context.
        Caller assumes lifetime management for the context.
        This must not return null (but may throw). */
    virtual IntContext* makeContext() = 0;
};

/** Simple context value. Makes its context as a clone of an existing one. */
class IntSimpleContextValue : public IntContextValue {
    IntContext* context;
 public:
    IntSimpleContextValue(IntContext* context);
    ~IntSimpleContextValue();

    // IntContextValue:
    IntContext* makeContext();

    // IntValue:
    string_t    toString(bool readable);
    void        store(IntTagNode&, Stream&);
    IntValue*   clone();

    static IntSimpleContextValue* createFromNewContext(IntContext* context);
};

/** Subroutine value. Value referring to a subroutine. This is used to
    implement the 'pushsub' opcode. */
class IntSubroutineValue : public IntValue {
    IntBCORef bco;
 public:
    IntSubroutineValue(IntBCORef bco);
    ~IntSubroutineValue();
    IntBCORef getBytecodeObject();

    string_t toString(bool readable);
    void store(IntTagNode& tag, Stream& aux);
    IntSubroutineValue* clone();
};

#endif
