/**
  *  \file int/exec.cc
  *  \brief Interpreter: Bytecode Process Execution
  */

#include <memory>
#include "int/exec.h"
#include "int/interror.h"
#include "int/binary.h"
#include "int/unary.h"
#include "int/context.h"
#include "int/compound.h"
#include "int/intglobal.h"

std::auto_ptr<IntContext> IntExecutionContext::global_context;

/*************************** IntExecutionFrame ***************************/

/** Construct stack frame.
    \param bco Byte code object associated with this frame (=code to execute) */
IntExecutionFrame::IntExecutionFrame(IntBCORef bco)
    : bco(bco),
      pc(0),
      context_sp(0),
      exception_sp(0)
{ }

/** Destroy stack frame. */
IntExecutionFrame::~IntExecutionFrame()
{ }

/************************** IntExecutionContext **************************/

/** Constructor. */
IntExecutionContext::IntExecutionContext()
{
    if (global_context.get())
        pushNewContext(global_context->clone());
}

/** Destructor. */
IntExecutionContext::~IntExecutionContext()
{ }

/** Push new frame. This frame will execute util termination, then execution
    will resume at the current frame. */
void
IntExecutionContext::pushNewFrame(IntExecutionFrame* frame)
{
    frame->context_sp   = contexts.size();
    frame->exception_sp = exceptions.size();
    frames.push_back(frame);
    contexts.push_back(new IntExecutionFrameContext(*frame));
}

/** Pop frame. Execution resumes at previous frame. */
void
IntExecutionContext::popFrame()
{
    ASSERT(frames.size() != 0);
    IntExecutionFrame* frame = frames.back();
    while (contexts.size() > frame->context_sp)
        contexts.pop_back();
    while (exceptions.size() > frame->exception_sp)
        exceptions.pop_back();
    frames.pop_back();
}

/** Push exception handler. If program triggers an error, execution resumes
    in the current frame at the specified program counter.
    \param pc Program counter */
void
IntExecutionContext::pushExceptionHandler(IntBytecodeObject::pc_t pc)
{
    IntExceptionHandler* e = new IntExceptionHandler;
    e->value_sp   = value_stack.size();
    e->context_sp = contexts.size();
    e->frame_sp   = frames.size();
    e->pc         = pc;
    exceptions.push_back(e);
}

/** Pop exception handler. Cancels a previous pushExceptionHandler. */
void
IntExecutionContext::popExceptionHandler()
{
    exceptions.pop_back();
}

/** Handle exception. Transfers control to the most recently pushed exception handler,
    if any, and have it process exception e. If there is no active handler, stops
    the bytecode process. */
void
IntExecutionContext::handleException(string_t e)
{
    if (exceptions.size()) {
        /* There is a user-specified exception handler */
        IntExceptionHandler* e = exceptions.back();

        /* Unwind stacks */
        while (value_stack.size() > e->value_sp)
            value_stack.pop();
        while (contexts.size() > e->context_sp)
            contexts.pop_back();
        while (frames.size() > e->frame_sp)
            frames.pop_back();

        /* FIXME: assign System.Error */

        /* Change program counter */
        frames.back()->pc = e->pc;

        /* Pop exception frame */
        exceptions.pop_back();
    } else {
        /* No user-specified exception handler, so reflect error to caller */
        // FIXME: this would normally set the process's state to "Terminated with Error"
        throw IntError(e);
    }
}

/** Push a new context. This one will be the first one used to look up variables. */
void
IntExecutionContext::pushNewContext(IntContext* ctx)
{
    contexts.push_back(ctx);
}

/** Copy contexts from a template. */
void
IntExecutionContext::pushContextsFrom(ptr_vector<IntContext>& ctxs)
{
    for (ptr_vector<IntContext>::iterator i = ctxs.begin(); i != ctxs.end(); ++i)
        pushNewContext((*i)->clone());
}

/** Pop context. Cancels a previous pushNewContext. */
void
IntExecutionContext::popContext()
{
    contexts.pop_back();
}

/** Run this process. */
void
IntExecutionContext::run()
{
    while (frames.size()) {
        try {
            executeInstruction();
        }
        catch (IntError& e) {
            handleException(e.what());
        }
    }
}

/** Execute a single instruction. */
void
IntExecutionContext::executeInstruction()
{
    IntExecutionFrame& f = *frames.back();

    // Run off end of code?
    if (f.pc >= f.bco->getNumInstructions()) {
        popFrame();
        return;
    }

    // Execute it
    const IntOpcode& op = (*f.bco)(f.pc++);
    switch (op.major) {
     case IntOpcode::maPush:
        /* Push value */
        switch (op.minor) {
         case IntOpcode::sNamedVariable:
            {
                int32_t index;
                if (IntContext* con = lookup(f.bco->getName(op.arg), index)) {
                    value_stack.pushNew(con->get(index));
                } else {
                    throw IntError::unknownIdentifier();
                }
            }
            break;
         case IntOpcode::sLocal:
            value_stack.push(f.local_values[op.arg]);
            break;
         case IntOpcode::sStatic:
            value_stack.push(frames.front()->local_values[op.arg]);
            break;
         case IntOpcode::sShared:
            value_stack.push(global_values[op.arg]);
            break;
         case IntOpcode::sLiteral:
            value_stack.push(f.bco->getLiteral(op.arg));
            break;
         case IntOpcode::sInteger:
            value_stack.pushNew(makeIntValue(int16_t(op.arg)));
            break;
         case IntOpcode::sBoolean:
            value_stack.pushNew(makeBoolValue(op.arg));
            break;
         case IntOpcode::sEmpty:
            value_stack.pushNew(0);
            break;
         case IntOpcode::sSubroutine:
            value_stack.pushNew(new IntSubroutineValue(f.bco->getSubroutine(op.arg)));
            break;
         default:
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maBinary:
        /* Binary operations */
        {
            checkStack(2);
            IntValue* a = value_stack.top(1);
            IntValue* b = value_stack.top(0);
            IntValue* result = executeBinaryOperation(op.minor, a, b);
            value_stack.pop();
            value_stack.pop();
            value_stack.pushNew(result);
        }
        break;

     case IntOpcode::maUnary:
        /* Unary operations */
        {
            checkStack(1);
            IntValue* result = executeUnaryOperation(op.minor, value_stack.top(0));
            value_stack.pop();
            value_stack.pushNew(result);
        }
        break;

     case IntOpcode::maJump:
        /* Jump instructions */
        if (op.minor & IntOpcode::jCatch) {
            // "catch"
            pushExceptionHandler(f.bco->getJumpTarget(op.minor, op.arg));
        } else {
            // "jump"
            switch (op.minor & IntOpcode::jAlways) {
             case IntOpcode::jAlways:
                // Jump always
                f.pc = f.bco->getJumpTarget(op.minor, op.arg);
                break;
             case 0:
                // Jump never
                break;
             default:
                // Jump on condition
                checkStack(1);
                int cond = getBoolValue(value_stack.top());
                int mask = (cond < 0 ? IntOpcode::jIfEmpty
                            : cond > 0 ? IntOpcode::jIfTrue
                            : IntOpcode::jIfFalse);
                if ((op.minor & mask) != 0)
                    f.pc = f.bco->getJumpTarget(op.minor, op.arg);
                break;
            }
            if (op.minor & IntOpcode::jPopAlways) {
                checkStack(1);
                value_stack.pop();
            }
        }
        break;

     case IntOpcode::maCallProc:
     case IntOpcode::maCallFunc:
        throw IntError::internalError("not implemented");

     case IntOpcode::maIndirect:
        /* Indirect call */
        switch (op.minor) {
         case IntOpcode::miIMCall:
         case IntOpcode::miIMLoad:
            checkStack(op.arg + 1);
            /* CALLIND nargs   rr:R:args      => rr
               PUSHIND nargs   rr:R:args      => rr:result */
            if (value_stack.top(op.arg) == 0) {
                /* Dereferencing null stays null, so just drop args */
                value_stack.pop(op.arg);
                if (op.minor == IntOpcode::miIMCall)
                    value_stack.pop();
            } else if (IntIndexableValue* iv = dynamic_cast<IntIndexableValue*>(value_stack.top(op.arg))) {
                /* We can load this */
                IntValue* value = iv->get(value_stack, value_stack.size() - op.arg, op.arg);

                /* Update stack */
                value_stack.pop(op.arg+1);
                if (op.minor == IntOpcode::miIMLoad)
                    value_stack.pushNew(value);
                else
                    delete value;
            } else {
                /* Error */
                throw IntError::typeError(IntError::tyIndexable);
            }
            break;
         case IntOpcode::miIMStore:
         case IntOpcode::miIMPop:
            /* STOREIND nargs   rr:val:R:args  => rr:val
               POPIND nargs     rr:val:R:args  => rr */
            checkStack(op.arg + 2);
            if (IntIndexableValue* iv = dynamic_cast<IntIndexableValue*>(value_stack.top(op.arg))) {
                /* We can assign this */
                IntValue* value = value_stack.top(op.arg+1);
                iv->set(value_stack, value_stack.size() - op.arg, op.arg, value);

                /* Update stack */
                value_stack.pop(op.arg+1);
                if (op.minor == IntOpcode::miIMPop) {
                    value_stack.pop();
                }
            } else {
                /* Error */
                throw IntError::typeError(IntError::tyIndexable);
            }
            break;
        }
        break;

     case IntOpcode::maStack:
        /* Stack operations */
        switch (op.minor) {
         case IntOpcode::miStackDup:
            checkStack(op.arg);
            value_stack.push(value_stack.top(op.arg));
            break;
         case IntOpcode::miStackDrop:
            checkStack(op.arg);
            for (uint32_t i = 0; i < op.arg; ++i)
                value_stack.pop();
            break;
         case IntOpcode::miStackSwap:
            checkStack(op.arg);
            value_stack.swap(op.arg, value_stack, 0);
            break;
         default:
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maPop:
     case IntOpcode::maStore:
        /* Pop or store into variable */
        checkStack(1);
        switch (op.minor) {
         case IntOpcode::sNamedVariable:
            {
                int32_t index;
                if (IntContext* con = lookup(f.bco->getName(op.arg), index)) {
                    con->set(index, value_stack.top());
                } else {
                    throw IntError::unknownIdentifier();
                }
            }
            break;
         case IntOpcode::sLocal:
            f.local_values.set(op.arg, value_stack.top());
            break;
         case IntOpcode::sStatic:
            frames.front()->local_values.set(op.arg, value_stack.top());
            break;
         case IntOpcode::sShared:
            global_values.set(op.arg, value_stack.top());
            break;
         default:
            handleInvalidOpcode();
        }
        if (op.major == IntOpcode::maPop) {
            value_stack.pop();
        }
        break;

     case IntOpcode::maMemref:
        switch (op.minor) {
         case IntOpcode::miIMCall:
         case IntOpcode::miIMLoad:
            /* Load/Evaluate TOS.field */
            checkStack(1);
            if (!value_stack.top()) {
                /* Dereferencing null stays null, no change to stack */
            } else if (IntContextValue* cv = dynamic_cast<IntContextValue*>(value_stack.top())) {
                /* It's a context */
                std::auto_ptr<IntContext> con(cv->makeContext());
                int32_t index = con->lookup(f.bco->getName(op.arg));
                if (index >= 0) {
                    /* Load permitted */
                    IntValue* v = con->get(index);
                    value_stack.pop();
                    if (op.minor == IntOpcode::miIMCall) {
                        delete v;
                    } else {
                        value_stack.pushNew(v);
                    }
                } else {
                    /* Name not found */
                    throw IntError::unknownIdentifier();
                }
            } else {
                /* Not a context */
                throw IntError::typeError(IntError::tyRecord);
            }
            break;

         case IntOpcode::miIMStore:
         case IntOpcode::miIMPop:
            /* Store/Pop into TOS.field */
            checkStack(2);
            if (IntContextValue* cv = dynamic_cast<IntContextValue*>(value_stack.top())) {
                /* It's a context */
                std::auto_ptr<IntContext> con(cv->makeContext());
                int32_t index = con->lookup(f.bco->getName(op.arg));
                if (index >= 0) {
                    /* Assignment permitted */
                    con->set(index, value_stack.top(1));
                } else {
                    /* Name not found */
                    throw IntError::unknownIdentifier();
                }
            } else {
                /* Not a context */
                throw IntError::typeError(IntError::tyRecord);
            }
            value_stack.pop();            // context
            if (op.minor == IntOpcode::miIMPop) {
                value_stack.pop();        // value
            }
            break;
         default:
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maDim:
        throw IntError::internalError("not implemented");

     case IntOpcode::maSpecial:
        switch (op.minor) {
         case IntOpcode::miSpecialUncatch:
            /* Cancel previous catch */
            popExceptionHandler();
            break;
         case IntOpcode::miSpecialReturn:
            /* Stop this frame, return to caller */
            throw IntError::internalError("not implemented");
         case IntOpcode::miSpecialWith:
            /* Add TOS to context stack */
            checkStack(1);
            if (IntContextValue* cv = dynamic_cast<IntContextValue*>(value_stack.top())) {
                /* Valid context */
                pushNewContext(cv->makeContext());
                value_stack.pop();
            } else {
                /* Not a context */
                throw IntError::typeError(IntError::tyRecord);
            }
            break;
         case IntOpcode::miSpecialEndWith:
            /* Cancel previous miSpecialWith */
            popContext();
            break;
         case IntOpcode::miSpecialFirstIndex:
            /* Start iteration */
            checkStack(1);
            if (IntIndexableValue* iv = dynamic_cast<IntIndexableValue*>(value_stack.top())) {
                if (IntContext* con = iv->makeFirstContext()) {
                    /* We have something to iterate over */
                    pushNewContext(con);
                    value_stack.pop();
                    value_stack.pushNew(makeBoolValue(1));
                } else {
                    /* Set exists but is empty */
                    value_stack.pop();
                    value_stack.push(0);
                }
            } else {
                /* This is not a set */
                throw IntError::typeError(IntError::tyIterable);
            }
            break;
         case IntOpcode::miSpecialNextIndex:
            /* Continue iteration */
            if (contexts.back()->next()) {
                value_stack.pushNew(makeBoolValue(1));
            } else {
                popContext();
                value_stack.push(0);
            }
            break;
         case IntOpcode::miSpecialEndIndex:
            /* Cancel iteration */
            popContext();
            break;
         default:
            handleInvalidOpcode();
        }
        break;
     default:
        handleInvalidOpcode();
    }
}

/** Set global context for execution. Every newly-created
    IntExecutionContext will have this context active. */
void
IntExecutionContext::setNewGlobalContext(IntContext* p)
{
    global_context.reset(p);
}

/** Perform name lookup.
    \param name  [in] Name to look up
    \param index [out] If found, the context's cookie for access to the name
    \return Context containing the name, 0 if none */
IntContext*
IntExecutionContext::lookup(string_t name, int32_t& index)
{
    for (uint32_t i = contexts.size(); i > 0; --i) {
        IntContext* c = contexts[i-1];
        int n = c->lookup(name);
        if (n >= 0) {
            index = n;
            return c;
        }
    }
    return 0;
}

/** Check that stack contains a number of values. */
void
IntExecutionContext::checkStack(uint32_t required)
{
    if (value_stack.size() < required)
        throw IntError::internalError("stack error");
}

void
IntExecutionContext::handleInvalidOpcode()
{
    throw IntError::internalError("invalid opcode");
}
