/**
  *  \file int/exec.h
  *  \brief Interpreter: Bytecode Process Execution
  */
#ifndef PCC_V2_INT_EXEC_H
#define PCC_V2_INT_EXEC_H

#include <memory>
#include "cpluslib/smartptr.h"
#include "int/bytecode.h"
#include "cpluslib/ptrvector.h"
#include "int/context.h"
#include "int/if/globalif.h"

/** Stack frame of an executing process. */
struct IntExecutionFrame {
    /** Bytecode object executing in this frame. */
    IntBCORef bco;

    /** Next instruction to execute. */
    IntBytecodeObject::pc_t pc;

    /** Local values (parameters, local variables). */
    IntDataSegment local_values;

    /** Local names (parameters, local variables). */
    IntVariableNames local_names;

    /** Top of context stack when this frame was opened. */
    uint32_t context_sp;

    /** Top of exception stack when this frame was opened. */
    uint32_t exception_sp;

    IntExecutionFrame(IntBCORef bco);
    ~IntExecutionFrame();
};

/** Exception handler. */
struct IntExceptionHandler {
    uint32_t                frame_sp;
    uint32_t                context_sp;
    uint32_t                value_sp;
    IntBytecodeObject::pc_t pc;
};

/** Execution Context (process). */
class IntExecutionContext {
 public:
    IntExecutionContext();
    ~IntExecutionContext();

    void pushNewFrame(IntExecutionFrame* frame);
    void popFrame();

    void pushExceptionHandler(IntBytecodeObject::pc_t pc);
    void popExceptionHandler();
    void handleException(string_t e);

    void pushNewContext(IntContext* ctx);
    void pushContextsFrom(ptr_vector<IntContext>& ctxs);
    void popContext();

    uint32_t getNumActiveFrames() const
        { return frames.size(); }

    IntValue* getResult()
        { return value_stack.top(); }

    void run();
    void executeInstruction();

    static void setNewGlobalContext(IntContext* p);

 private:
    IntContext* lookup(string_t name, int32_t& index);
    void checkStack(uint32_t required);
    void handleInvalidOpcode();

    /** All active stack frames. */
    ptr_vector<IntExecutionFrame> frames;

    /** All active contexts for name lookup. */
    ptr_vector<IntContext> contexts;

    /** All active exception handling contexts. */
    ptr_vector<IntExceptionHandler> exceptions;

    /** Value stack. */
    IntDataSegment value_stack;

    /** Global context template. \see setNewGlobalContext */
    static std::auto_ptr<IntContext> global_context;
};

#endif
