/**
  *  \file int/if/playerif.cc
  *  \brief Interpreter Game Interface: Player Properties
  */

#include "int/if/playerif.h"
#include "int/value.h"
#include "game/struct.h"
#include "game/racename.h"
#include "game/config.h"
#include "game/global.h"
#include "game/team.h"
#include "cpluslib/misc.h"
#include "int/if/ifutil.h"
#include "int/data.h"
#include "int/compound.h"
#include "int/interror.h"

/** Implementation of "Player(nn)".
    \param dseg Data segment containing function parameters
    \param index Index of function parameters
    \param nargs Number of function parameters
    \return value representing "Player(nn)" */
IntValue*
IFPlayerGet(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    int32_t pid;
    checkArgumentCount(nargs, 1);
    if (!checkIntArg(pid, dseg[index], 1, NUM_PLAYERS))
        return 0;
    return IntSimpleContextValue::createFromNewContext(new IntPlayerContext(pid));
}

/** Implementation of "ForEach Player".
    \return Context representing first player */
IntContext*
IFPlayerMake()
{
    return new IntPlayerContext(1);
}

/**************************** IntPlayerContext ***************************/

namespace {
    enum { PlayerProperty };
}

static const TIntMapping player_mapping[] = {
    // FIXME: missing PBPs
    { "BASES",               iplScoreBases,       PlayerProperty },
    { "PLANETS",             iplScorePlanets,     PlayerProperty },
    { "RACE",                iplFullName,         PlayerProperty },
    { "RACE$",               iplId,               PlayerProperty },
    { "RACE.ADJ",            iplAdjName,          PlayerProperty },
    { "RACE.ID",             iplRaceId,           PlayerProperty },
    { "RACE.MISSION",        iplMission,          PlayerProperty },
    { "RACE.SHORT",          iplShortName,        PlayerProperty },
    { "SCORE",               iplScore,            PlayerProperty },
    { "SHIPS",               iplScoreShips,       PlayerProperty },
    { "SHIPS.CAPITAL",       iplScoreCapital,     PlayerProperty },
    { "SHIPS.FREIGHTERS",    iplScoreFreighters,  PlayerProperty },
    { "TEAM",                iplTeam,             PlayerProperty },
};

IntPlayerContext::IntPlayerContext(int pid)
    : pid(pid)
{ }

IntPlayerContext::~IntPlayerContext()
{ }

int32_t
IntPlayerContext::lookup(string_t name)
{
    return lookupName(name, player_mapping, countof(player_mapping));
}

void
IntPlayerContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntPlayerContext::get(int32_t index)
{
    return getPlayerProperty(pid, IntPlayerProperty(player_mapping[index].index));
}

bool
IntPlayerContext::next()
{
    if (pid < NUM_PLAYERS) {
        ++pid;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntPlayerContext::clone()
{
    return new IntPlayerContext(pid);
}


/** Get player property.
    \param pid Player to query (out of range values are handled properly)
    \param ipl Player property to query */
IntValue*
getPlayerProperty(int pid, IntPlayerProperty ipl)
{
    /* Special case for out-of-range values */
    if (pid <= 0 || pid > NUM_PLAYERS) {
        if (ipl == iplId && pid >= 0)
            return makeIntValue(pid);
        else
            return 0;
    }

    /* Regular handling */
    const TGen& gen = getDisplayedTurn().getGen(::getPlayerId()).getGenRecord();
    switch (ipl) {
     case iplAdjName:
        /*+ Adjective name of this player. */
        return makeStringValue(player_racenames.getAdjName(pid));
     case iplFullName:
        /*+ Full name of this player. */
        return makeStringValue(player_racenames.getFullName(pid));
     case iplId:
        /*+ Player number. */
        return makeIntValue(pid);
     case iplMission:
        /*+ Special mission assigned to this player. */
        return makeIntValue(raceMissionId(pid));
     case iplRaceId:
        /*+ Race assigned to this player. */
        return makeIntValue(raceId(pid));
     case iplShortName:
        /*+ Short name of this player. */
        return makeStringValue(player_racenames.getShortName(pid));
     case iplTeam:
        /*+ Team this player is in. */
        return makeIntValue(getPlayerTeam(pid));
     case iplScoreBases:
        /*+ Number of bases this player has, according to score. */
        return makeIntValue(gen.scores[pid-1].num_bases);
     case iplScorePlanets:
        /*+ Number of planets this player has, according to score. */
        return makeIntValue(gen.scores[pid-1].num_planets);
     case iplScoreShips:
        /*+ Number of ships this player has, according to score. */
        return makeIntValue(gen.scores[pid-1].num_capital + gen.scores[pid-1].num_freighters);
     case iplScore:
        /*+ This player's Tim-score. */
        return makeIntValue(120L*gen.scores[pid-1].num_bases
                            + 10L*(gen.scores[pid-1].num_capital + gen.scores[pid-1].num_planets)
                            + gen.scores[pid-1].num_freighters);
     case iplScoreCapital:
        /*+ Number of capital ships this player has, according to score. */
        return makeIntValue(gen.scores[pid-1].num_capital);
     case iplScoreFreighters:
        /*+ Number of freighters this player has, according to score. */
        return makeIntValue(gen.scores[pid-1].num_freighters);
    }
    return 0;
}
