/**
  *  \file int/interror.cc
  *  \brief Interpreter: Error messages
  *
  *  The basic idea is to generate most error messages by explicit
  *  calls to constructor methods (e.g. IntError::rangeError()). This
  *  allows us to associate a little bit of meta-information with
  *  errors, and generate the final message when the error is
  *  presented to the user.
  *
  *  Note that the backtrace of an error can be obtained from the
  *  process' state and need not be stored with the error message.
  *
  *  TODO: if a user process does
  *      Try / ... / Else / Throw System.Error / EndTry
  *  it would lose this meta-information because the error message is
  *  converted to a string inbetween.
  */

#include "int/interror.h"

/******************************** IntError *******************************/

IntError::IntError(const char* error)
    : error(error)
{ }

IntError::IntError(string_t error)
    : error(error)
{ }

IntError::~IntError() throw()
{ }

const char*
IntError::what() const throw()
{
    return error.c_str();
}

/** Generate an "unknown identifier" error. */
IntError
IntError::unknownIdentifier()
{
    // The identifier can be found in the executing BCO's current instruction.
    return IntError("Unknown identifier");
}

/** Generate a "type mismatch" error. */
IntError
IntError::typeError(ExpectedType expectedType)
{
    // FIXME
    static const char*const suffix[] = {
        "",
        ", expecting integer",
        ", expecting bool",
        ", expecting number",
        ", expecting string",
        ", expecting list or function",
        ", expecting list",
        ", expecting record",
    };
    return IntError(string_t("Type error") + suffix[expectedType]);
}

/** Generate an internal error. */
IntError
IntError::internalError(const char* msg)
{
    // FIXME
    return IntError(string_t("Internal error: ") + msg);
}

/** Generate an "object not serializable" error. This error is
    generated when an object incapable of serializing is asked to
    serialize (IntValue::store), which normally only happens when a
    process suspends. */
IntError
IntError::notSerializable()
{
    return IntError("Not suspendable");
}

/** Generate a "not assignable" error. This error is generated when an
    assignment is attempted to a value which is not assignable
    (usually, a read-only property). */
IntError
IntError::notAssignable()
{
    return IntError("Attempt to assign read-only value");
}

/** Generate a range error. */
IntError
IntError::rangeError()
{
    return IntError("Range error");
}
