/**
  *  \file int/opcode.cc
  *  \brief Interpreter: Opcodes
  */

#include <ostream>
#include "int/opcode.h"
#include "cpluslib/misc.h"
#include "int/unary.h"
#include "int/binary.h"

static const char*const scopes[]   = { "var", "loc", "top", "glob", "lit", "int", "bool", "e", "sub" };
static const char*const ims[]      = { "call", "load", "store", "pop" };
static const char*const stack[]    = { "dup", "drop", "swap" };
static const char*const specials[] = { "uncatch", "return", "with", "endwith", "firstindex", "nextindex", "endindex" };

static const char*
formatEnum(uint8_t minor, const char*const* values, std::size_t count)
{
    return minor < count ? values[minor] : "?";
}

static const char*
formatScope(uint8_t minor)
{
    return formatEnum(minor, scopes, countof(scopes));
}

static const char*
formatIM(uint8_t minor)
{
    return formatEnum(minor, ims, countof(ims));
}

/** Convert opcode to user-readable text form for disassembly. */
std::ostream&
operator<<(std::ostream& out, const IntOpcode& opcode)
{
    switch (opcode.major) {
     case IntOpcode::maPush:
        out << "push" << formatScope(opcode.minor) << "\t" << opcode.arg;
        break;
     case IntOpcode::maBinary:
        out << "b" << getBinaryName(opcode.minor);
        break;
     case IntOpcode::maUnary:
        out << "u" << getUnaryName(opcode.minor);
        break;
     case IntOpcode::maJump:
        {
            uint8_t flags = opcode.minor & ~IntOpcode::jSymbolic;
            if (flags == 0) {
                out << "label";
            } else if (flags < IntOpcode::jCatch) {
                out << "j";
                if ((flags & IntOpcode::jAlways) != IntOpcode::jAlways) {
                    if (flags & IntOpcode::jIfTrue)
                        out << "t";
                    if (flags & IntOpcode::jIfFalse)
                        out << "f";
                    if (flags & IntOpcode::jIfEmpty)
                        out << "e";
                    if ((flags & IntOpcode::jAlways) == 0)
                        out << "never"; /* might occur as jneverp */
                }
                if (flags & IntOpcode::jPopAlways)
                    out << "p";
            } else if (flags == IntOpcode::jCatch) {
                out << "catch";
            } else {
                out << "jump-" << flags;
            }

            if (opcode.minor & IntOpcode::jSymbolic)
                out << "\tsym" << opcode.arg;
            else
                out << "\t#" << opcode.arg;
        }
        break;
     case IntOpcode::maCallProc:
        out << "callp\t" << opcode.arg << "(" << int(opcode.minor) << ")";
        break;
     case IntOpcode::maCallFunc:
        out << "callf\t" << opcode.arg << "(" << int(opcode.minor) << ")";
        break;
     case IntOpcode::maIndirect:
        out << formatIM(opcode.minor) << "ind\t" << opcode.arg;
        break;
     case IntOpcode::maStack:
        out << formatEnum(opcode.minor, stack, countof(stack)) << "\t" << opcode.arg;
        break;
     case IntOpcode::maPop:
        out << "pop" << formatScope(opcode.minor) << "\t" << opcode.arg;
        break;
     case IntOpcode::maStore:
        out << "store" << formatScope(opcode.minor) << "\t" << opcode.arg;
        break;
     case IntOpcode::maMemref:
        out << formatIM(opcode.minor) << "mem\t" << opcode.arg;
        break;
     case IntOpcode::maDim:
        out << "dim" << formatScope(opcode.minor) << "\t" << opcode.arg;
        break;
     case IntOpcode::maSpecial:
        out << formatEnum(opcode.minor, specials, countof(specials)) << "\t" << opcode.arg;
        break;
     default:
        out << "unk-" << int(opcode.major) << "-" << int(opcode.minor) << "\t" << opcode.arg;
    };

    return out;
}
